@@ -82,6 +82,7 @@ fatal_1350(cPersistentObject *self, const char *caller, const char *detail)
 #endif
 
 static void ghostify(cPersistentObject*);
+static PyObject * pickle_slotnames(PyTypeObject *cls);
 
 static PyObject * convert_name(PyObject *name);
 
@@ -150,7 +151,7 @@ accessed(cPersistentObject *self)
 static void
 ghostify(cPersistentObject *self)
 {
-    PyObject **dictptr;
+    PyObject **dictptr, *slotnames;
 
     /* are we already a ghost? */
     if (self->state == cPersistent_GHOST_STATE)
@@ -180,6 +181,8 @@ ghostify(cPersistentObject *self)
         _estimated_size_in_bytes(self->estimated_size);
     ring_del(&self->ring);
     self->state = cPersistent_GHOST_STATE;
+
+    /* clear __dict__ */
     dictptr = _PyObject_GetDictPtr((PyObject *)self);
     if (dictptr && *dictptr)
     {
@@ -187,6 +190,53 @@ ghostify(cPersistentObject *self)
         *dictptr = NULL;
     }
 
+    /* clear all slots besides _p_*
+     * ( for backward-compatibility reason we do this only if class does not
+     *   override __new__ ) */
+    if (Py_TYPE(self)->tp_new == Pertype.tp_new)
+    {
+        slotnames = pickle_slotnames(Py_TYPE(self));
+        if (slotnames && slotnames != Py_None)
+        {
+            int i;
+
+            for (i = 0; i < PyList_GET_SIZE(slotnames); i++)
+            {
+                PyObject *name;
+                char *cname;
+                int is_special;
+
+                name = PyList_GET_ITEM(slotnames, i);
+#ifdef PY3K
+                if (PyUnicode_Check(name))
+                {
+                    PyObject *converted = convert_name(name);
+                    cname = PyBytes_AS_STRING(converted);
+#else
+                if (PyBytes_Check(name))
+                {
+                    cname = PyBytes_AS_STRING(name);
+#endif
+                    is_special = !strncmp(cname, "_p_", 3);
+#ifdef PY3K
+                    Py_DECREF(converted);
+#endif
+                    if (is_special) /* skip persistent */
+                    {
+                        continue;
+                    }
+                }
+
+                /* NOTE: this skips our delattr hook */
+                if (PyObject_GenericSetAttr((PyObject *)self, name, NULL) < 0)
+                    /* delattr of non-set slot will raise AttributeError - we
+                     * simply ignore. */
+                    PyErr_Clear();
+            }
+        }
+        Py_XDECREF(slotnames);
+    }
+
     /* We remove the reference to the just ghosted object that the ring
     * holds.  Note that the dictionary of oids->objects has an uncounted
     * reference, so if the ring's reference was the only one, this frees
@@ -270,6 +320,8 @@ Per__p_deactivate(cPersistentObject *self)
             called directly. Methods that override this need to
             do the same! */
         ghostify(self);
+        if (PyErr_Occurred())
+            return NULL;
     }
 
     Py_INCREF(Py_None);
@@ -298,6 +350,8 @@ Per__p_invalidate(cPersistentObject *self)
         if (Per_set_changed(self, NULL) < 0)
             return NULL;
         ghostify(self);
+        if (PyErr_Occurred())
+            return NULL;
     }
     Py_INCREF(Py_None);
     return Py_None;
@@ -534,6 +588,8 @@ pickle___setstate__(PyObject *self, PyObject *state)
     if (state != Py_None)
     {
         PyObject **dict;
+        PyObject *d_key, *d_value;
+        Py_ssize_t i;
 
         dict = _PyObject_GetDictPtr(self);
         
@@ -552,8 +608,19 @@ pickle___setstate__(PyObject *self, PyObject *state)
         }
 
         PyDict_Clear(*dict);
-        if (PyDict_Update(*dict, state) < 0)
-            return NULL;
+
+        i = 0;
+        while (PyDict_Next(state, &i, &d_key, &d_value)) {
+            /* normally the keys for instance attributes are
+               interned.  we should try to do that here. */
+            if (NATIVE_CHECK_EXACT(d_key)) {
+                Py_INCREF(d_key);
+                INTERN_INPLACE(&d_key);
+                Py_DECREF(d_key);
+            }
+            if (PyObject_SetItem(*dict, d_key, d_value) < 0)
+                return NULL;
+        }
     }
 
     if (slots && pickle_setattrs_from_dict(self, slots) < 0)
